/**
 * imageMaskCropper.
 * jQuery library that gives you ability to move images across container using mask as well.
 *
 * @todo   make it modular.
 * @author Miłosz Sobczak <milosz.sobczak@pixers.pl>
 * @return {object} imageMaskCropper plugin
 */
;(function (factory) { 
    if (typeof define === 'function' && define.amd) {
        // AMD. Register as an anonymous module. 
        define(['jquery'], factory); 
    }
    else if (typeof exports === 'object') {
        // Node/CommonJS 
        factory(require('jquery')); 
    }
    else { 
        // Browser globals 
        factory(window.jQuery || window.Zepto); 
    } 
 }(function($) {
    /**
     * Loads image and returns promise.
     * 
     * @param  {string} src srouce of image
     * @return {promise}
     */
    var loadImage = function (src) {
        var defer = $.Deferred(),
            $img = $('<img>');

        $img.attr('src', src).on('load', function () {
            defer.resolve({
                width: $img.get(0).width,
                height: $img.get(0).height
            });
        }).each(function(){
          if (this.complete) {
            $(this).trigger('load');
          }
        });

        return defer;
    };

    /**
     * Gets opposite dimension.
     * 
     * @param  {string}
     * @return {string}
     */
    var getOpositeDimension = function (dimension) {
        return (dimension === 'height') ? 'width' : 'height';
    }
    
    /* <--plugin core-->  */
    var elementObject = {
        width: null,
        height: null,
        originalWidth: null,
        originalHeight: null,
        $element: null
    };
    /**
     * Main imageMaskCropper class.
     *
     * @param  {object} options  plugin options
     * @param  {object} $element jquery main image element
     * @return {object} self instance
     */
    var imageMaskCropper = function (options, $element) {
        this.eventsInitialized = false;

        this.options = options;
        this.element = Object.create(elementObject);
        this.element.$element = $element

        this.imageMain = Object.create(elementObject);
        this.imageMask = Object.create(elementObject);
        this.container = Object.create(elementObject);
        this.window = Object.create(elementObject);

        this.cropZone = Object.create(elementObject);

        this.isRunning = false;
        this._isUser = false;

        this._init()
        return this;
    }
    imageMaskCropper.prototype = {
        /**
         * Cropper initialization.
         * @return {promise|void}
         */
        _init: function () {
            var self = this;

            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t initialize.');
            }
            this.isRunning = true;

            if (this._isContainerExistingInDom()) {
                this.container.$element.addClass(this.options.cropperContainerClassLoading);
            }

            loadImage(this.options.imageMain).then(function (size) {
                self.imageMain.originalHeight = size.height;
                self.imageMain.originalWidth = size.width;

                if (self._isMaskDefinedInOptions()) {
                    return loadImage(self.options.imageMask).then(function (size) {
                        self.imageMask.originalHeight = size.height;
                        self.imageMask.originalWidth = size.width
                        self._refreshContainer();
                    });
                }
                self._refreshContainer(); 
            });
        },
        /**
         * Changes imageMain src.
         * @param  {string} src
         * @return {void|boolean}
         */
        changeImageMain: function (src) {
            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t change imageMain.');
            }
            if (typeof src !== 'string' || (this.imageMain.$element !== null && this.imageMain.$element.attr('src') === src)) {
                return false;
            }
            this.options.imageMain = src;
            this._init();
        },
        /**
         * Changes imageMask and maskOptions at once.
         * @return {void|boolean}
         */
        changeImageMaskAndItsOptions: function (src, maskOptions) {
            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t change imageMask and its options.');
            }
            if ((typeof src !== 'string' || (this.imageMask.$element !== null && this.imageMask.$element.attr('src') === src)) || typeof maskOptions !== 'object') {
                return false;
            }
            this.options.imageMask = src;
            this.options.maskOptions = $.extend(true, this.options.maskOptions, maskOptions);
            this._init();
        },
        /**
         * Changes imageMask src.
         * @param  {string} src
         * @return {void|boolean}
         */
        changeImageMask: function (src) {
            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t change imageMask.');
            }
            if (typeof src !== 'string' || (this.imageMask.$element !== null && this.imageMask.$element.attr('src') === src)) {
                return false;
            }
            this.options.imageMask = src;
            this._init();
        },
        /**
         * Changes maskOptions.
         * @param  {object} maskOptions
         * @return {void|boolean}
         */
        changeMaskOptions: function (maskOptions) {
            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t change maskOptions.');
            }
            if (typeof maskOptions !== 'object') {
                return false;
            }
            var shouldInit = 
              ((maskOptions.left && (this.options.maskOptions.left !== maskOptions.left)) ||
              (maskOptions.top && (this.options.maskOptions.top !== maskOptions.top))) ? true : false;

            this.options.maskOptions = $.extend(true, this.options.maskOptions, maskOptions);

            if (shouldInit) {
              return this.init();
            }
            this.setPositionCoords(this._getPositionCoordsValid(this.options.maskOptions));
        },
        /**
         * Gets coordinatates of crop.
         * 
         * @throws {error} if app is running
         * @return {object} crop dimensions in percentage.
         */
        getCropPosition: function () {
            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t get crop position.');
            }
            var self = this,
                heightPercent = 0,
                widthPercent = 0,
                leftPercent = 0,
                topPercent = 0;

            if (this._isMaskDefinedInOptions()) {
                leftPercent = ((self.options.maskOptions.left - (self.imageMain.$element.position().left) / self.container.width * 100) / 100 * self.container.width) / self.imageMain.width * 100,
                topPercent = ((self.options.maskOptions.top - (self.imageMain.$element.position().top) / self.container.height * 100) / 100 * self.container.height / self.imageMain.height * 100);
                heightPercent = ((self.options.maskOptions.height * self.container.height / 100) / (self.container.height - (self.container.height - self.imageMain.height)) * 100);
                widthPercent = ((self.options.maskOptions.width * self.container.width / 100) / (self.container.width - (self.container.width - self.imageMain.width)) * 100);
            }
            else {
                leftPercent = self.cropZone.$element.find('[class*="' + self.options.cropZoneClass + '-selection"]').position().left / self.container.width * 100,
                topPercent = self.cropZone.$element.find('[class*="' + self.options.cropZoneClass + '-selection"]').position().top / self.container.height * 100;
                heightPercent = self.cropZone.$element.find('[class*="' + self.options.cropZoneClass + '-selection"]').height() / self.container.height * 100;
                widthPercent = self.cropZone.$element.find('[class*="' + self.options.cropZoneClass + '-selection"]').width() / self.container.width * 100;
            }

            return {
                left: Number(Math.max(Math.min(leftPercent, 100), 0).toFixed(8)),
                top: Number(Math.max(Math.min(topPercent, 100), 0).toFixed(8)),
                height: Number(Math.max(Math.min(heightPercent, 100), 0).toFixed(8)),
                width: Number(Math.max(Math.min(widthPercent, 100), 0).toFixed(8))
            };
        },
        /**
         * Sets current Crop position.
         * @param  {object} diffPercentCoords {x: percentValue, y: percentValue}
         * @todo add posibility to move without mask as well
         * @return {void}
         */
        setPositionCoords: function (diffPercentCoords) {
            var imgPercent = null,
                maskPercent = null;

            this._setPositionCoordsY(diffPercentCoords.y);
            this._setPositionCoordsX(diffPercentCoords.x);

            if ($.isNumeric(diffPercentCoords.y)) {
                if (this._isMaskDefinedInOptions()) {
                    imgPercent = diffPercentCoords.y * this.imageMain.height / 100;
                    maskPercent = this.options.maskOptions.top * this.container.height / 100;
                    this._setPositionTopOfImageMain(-((imgPercent - maskPercent) / this.container.height * 100));
                }
                else {
                    this._setPositionOfCropZone();
                }
            }
            if ($.isNumeric(diffPercentCoords.x)) {
                if (this._isMaskDefinedInOptions()) {
                    imgPercent = diffPercentCoords.x * this.imageMain.width / 100;
                    maskPercent = this.options.maskOptions.left * this.container.width / 100;
                    this._setPositionLeftOfImageMain(-((imgPercent - maskPercent) / this.container.width * 100));   
                }
                else {
                    this._setPositionOfCropZone();
                }
            }

            this._triggerMoveFunc();
        },
        /**
         * Setter.
         * Sets maskOptionsHeight and adjust coords to it.
         * @param {float} height
         * @return {void}
         */
        setMaskOptionsHeight: function (height) {
            this._setMaskOptionsHeight(height);
            this.setPositionCoords(this._getPositionCoordsValid(this.options.maskOptions));
        },
        /**
         * Setter.
         * Sets maskOptionsWidth and adjust coords to it.
         * @param {float} width
         * @return {void}
         */
        setMaskOptionsWidth: function (width) {
            this._setMaskOptionsWidth(width)
            this.setPositionCoords(this._getPositionCoordsValid(this.options.maskOptions));
        },
        /**
         * Enables crop plugin option.
         * @return {void|boolean}
         */
        enableCrop: function () {
            if (this.options.enableCrop === true) {
                return false;
            }
            this.options.enableCrop = true;
            this._updateContainerClass();
        },
        /**
         * Disables crop plugin option.
         * @return {void|boolean}
         */
        disableCrop: function () {
            if (this.options.enableCrop === false) {
                return false;
            }
            this.options.enableCrop = false;
            this._updateContainerClass();
        },
        /**
         * Gets valid position coords based on mask options.
         * 
         * @param  {object} maskOptions
         * @return {object} crop positions
         */
        _getPositionCoordsValid: function (maskOptions) {
          var xDiff = Math.max((this.options.positionCoords.x + maskOptions.width) - 100, 0),
              yDiff = Math.max((this.options.positionCoords.y + maskOptions.height) - 100, 0);

          return {
              x: this.options.positionCoords.x + xDiff,
              y: this.options.positionCoords.y + yDiff
          };
        },
        /**
         * Triggers callback func when crop is moved.
         * @return {void}
         */
        _triggerMoveFunc: function () {
            var self = this;

            if (typeof self.options.onMoveCrop !== 'function') {
                return;
            }

            self.options.onMoveCrop.call(self, {
                coords: self.getCropPosition(),
                isUser: self._isUser
            });
        },
        /**
         * Refresh whole scene based on configuration.
         * @return {void}
         */
        _refreshContainer: function () {
            var self = this,
                onInitialized = function () {
                    self.container.$element.removeClass(self.options.cropperContainerClassLoading);
                    if (typeof self.options.onInitialized === 'function') {
                        self.options.onInitialized.call(self, {
                            coords: self.getCropPosition(),
                            imageMain: self.options.imageMain
                        });
                    }
                };

            this._isUser = false;

            if (this._isContainerExistingInDom() === false) {
                this._createContainerDOMElement();
            }
            if (this._isImageMaskExistingInDom() === true && this._isMaskDefinedInOptions() === false) {
                this._removeDomImageMask();
                this._removeImageMainStyles();
            }
            this._updateContainerClass();
            this._hideInitializingElement();

            if (this._isImageMainExistingInDom() === false) {
                this._createDomImageMain();
            }
            else {
                this._updateDomImageMain();
            }
            if (this._isMaskDefinedInOptions() === true) {
                if (this._isCropZoneExistingInDom() == true) {
                    this._removeDomCropZone();
                }
                if (this._isImageMaskExistingInDom() === false) {
                    this._createDomImageMask();
                }
                else {
                    this._updateDomImageMask();
                }
            }
            else if (this._isCropZoneExistingInDom() == false){
                this._createDomCropZone();
            }

            setTimeout(function () {
                self.imageMain.$element.removeAttr('style');

                if (self._isMaskDefinedInOptions() === true) {
                    self._setPositionOfImageMain();
                }
                else {
                    self._setPositionOfCropZone();
                }
                self._setContainerCssHeight('auto');
                if (self.eventsInitialized === false) {
                    self._initEvents();
                }
                self._udpateElementsSizes();
                self._updateCacheSizeOfWindow();
                self.isRunning = false;
                onInitialized();
            }, 0);
        },
        /**
         * Events initialization.
         * @return {void}
         */
        _initEvents: function () {
            this._onPluginMoveCurrentCrop();
            this._onWindowsResize();
            this._onMoveImageOrCropZone();
        },
        /**
         * Events listening to move image in mask.
         * @return {void}
         */
        _onPluginMoveCurrentCrop: function () {
            var self = this;
            this.element.$element.on('imageMaskCropper:moveCurrentCrop', function (e,  diffPercentCoords) {
                self.setPositionCoords(diffPercentCoords);

                self._triggerMoveFunc();
            });
        },
        /**
         * Updates cache sizes for Elements.
         * @return {void}
         */
        _udpateElementsSizes: function () {
            if (this.options.isMobile) {
                this._setMaxWidthOfContainer();
            }

            this._updateCacheSizeOfWindow();
            this._updateCacheSizeOfContainer();
            this._updateCacheSizeOfElement();
            this._updateCacheSizeOfImageMain();
            if (this._isMaskDefinedInOptions() == true) {
                this._updateCacheSizeOfImageMask();
            }
        },
        /**
         * On window resize event.
         * @return {void}
         */
        _onWindowsResize: function () {
            var self = this,
                _resizeThrottle,
                _throttleTime = 250,
                minDiffToResize = 40,
                sumOfAbsHeightDiff = 0;

            $(window).resize(function () {
                var absDiff = Math.abs(self.window.height - $(window).height());

                sumOfAbsHeightDiff = sumOfAbsHeightDiff + absDiff;
                self._updateCacheSizeOfWindow();
                
                if (typeof _resizeThrottle !== 'undefined') {
                    clearTimeout(_resizeThrottle);
                }

                _resizeThrottle = setTimeout(function () {
                    if (sumOfAbsHeightDiff > minDiffToResize) {
                        self._udpateElementsSizes();
                    }
                    sumOfAbsHeightDiff = 0;
                }, _throttleTime);
            });
        },
        /**
         * On move events.
         * @return {void}
         */
        _onMoveImageOrCropZone: function () {
            var self = this,
                startMousePos = null,
                startPosCSS = null,
                $elToMove = (function () {
                    if (self._isMaskDefinedInOptions()) {
                        return self.imageMain.$element;
                    }
                    return self.cropZone.$element.find('[class*="' + self.options.cropZoneClass + '-selection"]');
                })(),
                eventsHandler = {
                    down: function (e) {
                        if (self.options.enableCrop === true) {
                            e.preventDefault();

                            startMousePos = {
                                x: e.pageX || (function () {
                                    if (typeof e.originalEvent !== 'undefined') {
                                        return e.originalEvent.touches[0].pageX;
                                    }
                                    return null;
                                })(),
                                y: e.pageY || (function () {
                                    if (typeof e.originalEvent !== 'undefined') {
                                        return e.originalEvent.touches[0].pageY;
                                    }
                                    return null;
                                })(),
                            };

                            startPosCSS = {
                                top: $elToMove.position().top,
                                left: $elToMove.position().left,
                            };
                        }
                    },
                    move: function (e) {
                        var diff = {
                                x: null,
                                y: null
                            },
                            coords = {};

                        if (startMousePos === null || startPosCSS === null) {
                            return;
                        }

                        coords = self._getCoordinatesFromEvent(e);

                        if (coords.x !== null && coords.y !== null) {
                            diff.x = startMousePos.x - coords.x;
                            diff.y = startMousePos.y - coords.y;

                            self._moveCropZoneOrImage (startPosCSS, diff);
                        }
                    },
                    end: function (e) {
                        startMousePos = null;
                        startPosCSS = null;
                    }
                };

            if (this.options.isMobile) {
                if (window.PointerEvent) {
                    self.container.$element.on('pointerdown', eventsHandler.down);
                    self.container.$element.on('pointermove', eventsHandler.move);
                    self.container.$element.on('pointerup', eventsHandler.end);
                }
                else {
                    self.container.$element.on('touchstart', eventsHandler.down);
                    self.container.$element.on('touchmove', eventsHandler.move);
                    self.container.$element.on('touchend', eventsHandler.end);
                }
            }
            else {
                self.container.$element.on('mousedown', eventsHandler.down);
                $(document).on('mousemove', eventsHandler.move);
                $(document).on('mouseup mouseleave', eventsHandler.end);
            }
        },
        /**
         * Gets coordinates from mouse/touch event.
         * 
         * @param  {object} e event
         * @return {object}
         */
        _getCoordinatesFromEvent: function (e) {
          return {
              x: (function (pageX) {
                  if (pageX) {
                      return pageX;
                  }
                  if (e.originalEvent.touches.length > 0) {
                      return e.originalEvent.touches[0].pageX;
                  }
                  return null;
              })(e.pageX),
              y: (function (pageY) {
                  if (pageY) {
                      return pageY;
                  }
                  if (e.originalEvent.touches.length > 0) {
                      return e.originalEvent.touches[0].pageY;
                  }
                  return null;
              })(e.pageY)
          };
        },
        /**
         * Handles movement of image/crop.
         * @param  {object} startPosCSS starting positiin
         * @param  {object} diff        difference between current and starting
         * @return {void}
         */
        _moveCropZoneOrImage: function (startPosCSS, diff) {
            this._isUser = true;

            if (this.isRunning === true) {
                throw new Error('Plugin is still running. Can\'t move.');
            }
            if (this._isMaskDefinedInOptions()) {
                return this._moveImageInMask(startPosCSS, diff);
            }
            return this._moveCropZone(startPosCSS, diff);
        },
        /**
         * Moves images across imageMask.
         * @param  {object} startPosCSS starting positiin
         * @param  {object} diffCoords        difference between current and starting
         * @return {void}
         */
        _moveImageInMask: function (startPosCSS, diffCoords) {
            var self = this,
                countedValue = null;

            if (this.imageMain.adjustedSize === 'width') {
                countedValue = this._getTopPercentOfContainer(startPosCSS.top - diffCoords.y);
                this._setPositionTopOfImageMain(countedValue);
            }
            else {
                countedValue = this._getLeftPercentOfContainer(startPosCSS.left - diffCoords.x);
                this._setPositionLeftOfImageMain(countedValue);
            }

            this._triggerMoveFunc();
        },
        /**
         * Moves crop above imageMain.
         * @param  {object} startPosCSS starting positiin
         * @param  {object} diffCoords  difference between current and starting
         * @return {void}
         */
        _moveCropZone: function (startPosCSS, diffCoords) {
            var self = this;

            this._setPositionCoordsX(this._getLeftPercentOfContainer(startPosCSS.left - diffCoords.x));
            this._setPositionCoordsY(this._getTopPercentOfContainer(startPosCSS.top - diffCoords.y));
            
            this._setPositionOfCropZone();
            this._triggerMoveFunc();
        },
        /**
         * Gets percent of container's height.
         * @param  {float} value value in pixels
         * @return {float}
         */
        _getTopPercentOfContainer: function (value) {
            return value / this.container.height * 100;
        },
        /**
         * Gets percent of container's width.
         * @param  {float} value value in pixels
         * @return {float}
         */
        _getLeftPercentOfContainer: function (value) {
            return value / this.container.width * 100;
        },
        /**
         * Hides $element.
         * @return {void}
         */
        _hideInitializingElement: function () {
            this.element.$element.hide();
        },
        /**
         * Creates the container div in cropperContainerClass.
         * @return {void}
         */
        _createContainerDOMElement: function () {
            this.container.$element = $('<div/>').addClass(this.options.cropperContainerClass + ' ' + this.options.cropperContainerClassLoading).insertAfter(this.element.$element);
        },
        /**
         * Updates container class.
         * @return {void}
         */
        _updateContainerClass: function () {
            if (this._isMaskDefinedInOptions()) {
                this.container.$element.addClass(this.options.cropperContainerClassHasMask);
            }
            else {
                this.container.$element.removeClass(this.options.cropperContainerClassHasMask);
            }
            if (this.options.enableCrop === true) {
                this.container.$element.addClass(this.options.cropperContainerClassEnabledCrop);   
            }
            else {
                this.container.$element.removeClass(this.options.cropperContainerClassEnabledCrop);      
            }
        },
        /**
         * Updates src of imageMain based on plugin options.
         * @return {void}
         */
        _updateDomImageMain: function () {
            if (this.options.imageMain !== this.imageMain.$element.attr('src')) {
                this.imageMain.$element.attr('src', this.options.imageMain);
                this._updateCacheSizeOfImageMain();
            }
        },
        /**
         * Updates src of imageMask based on plugin options.
         * @return {void}
         */
        _updateDomImageMask: function () {
            if (this.options.imageMask !== this.imageMask.$element.attr('src')) {
                this.imageMask.$element.attr('src', this.options.imageMask);
                this._updateCacheSizeOfImageMask();
            }
        },
        /**
         * Sets position of cropZone.
         */
        _setPositionOfCropZone: function () {
            this._adjustImageMainToMaskOrCropZone();
            this._updateCSSForCropZoneSelection();
            this._updateCSSForCropZoneTop();
            this._updateCSSForCropZoneRight();
            this._updateCSSForCropZoneBottom();
            this._updateCSSForCropZoneLeft();
        },
        /**
         * Adds CSS for cropzone selection element.
         * @return {void}
         */
        _updateCSSForCropZoneSelection: function() {
            var css = {
                top: this.options.positionCoords.y + '%',
                left: this.options.positionCoords.x + '%',
                width: this.options.maskOptions.width + '%',
                height: this.options.maskOptions.height + '%',
            };

            this.cropZone.$element.find('[class*="' + this.options.cropZoneClass + '-selection"]').css(css);
        },
        /**
         * Adds CSS for cropzone top element.
         * @return {void}
         */
        _updateCSSForCropZoneTop: function() {
            var options = this.options,
                css = {
                    height: options.positionCoords.y + '%',
                };

            this.cropZone.$element.find('[class*="' + options.cropZoneClass + '-top"]').css(css);
        },
        /**
         * Adds CSS for cropzone right element.
         * @return {void}
         */
        _updateCSSForCropZoneRight: function() {
            var options = this.options,
                css = {
                    top: options.positionCoords.y + '%',
                    height: options.maskOptions.height + '%',
                    width: (100 - (options.positionCoords.x + options.maskOptions.width)) + '%'
                };

            this.cropZone.$element.find('[class*="' + options.cropZoneClass + '-right"]').css(css);
        },
        /**
         * Adds CSS for cropzone bottom element.
         * @return {void}
         */
        _updateCSSForCropZoneBottom: function() {
            var options = this.options,
                css = {
                    height: (100 - (options.positionCoords.y + options.maskOptions.height)) + '%',
                };

            this.cropZone.$element.find('[class*="' + options.cropZoneClass + '-bottom"]').css(css);
        },
        /**
         * Adds CSS for cropzone left element.
         * @return {void}
         */
        _updateCSSForCropZoneLeft: function() {
            var options = this.options,
                css = {
                    top: options.positionCoords.y + '%',
                    height: options.maskOptions.height + '%',
                    width: options.positionCoords.x + '%'
                };

            this.cropZone.$element.find('[class*="' + options.cropZoneClass + '-left"]').css(css);
        },
        /**
         * Sets position of imageMain in imageMask.
         * @return {void}
         */
        _setPositionOfImageMain: function () {
            this._adjustImageMainToMaskOrCropZone();
            if (this.imageMain.adjustedSize === 'width' || this.imageMain.adjustedSize === 'height') {
                this._centerImageMainInMask();
            }
        },
        /**
         * Sets top positon of image using given param.
         * @param {float} topValue percentage value
         * @param {boolean} stop 
         */
        _setPositionTopOfImageMain: function (topValue, stop) {
            var bottomPercentInOption = 0;
            this.imageMain.$element.css('top', topValue + '%');

            if (stop === true) {
                return false;
            }

            if (this._getPercentTopOfImageMain() > this.options.maskOptions.top) {
                this._setPositionTopOfImageMain(this.options.maskOptions.top, true);
            }

            bottomPercentInOption = (100 - (this.options.maskOptions.top + this.options.maskOptions.height));

            if (this._getPercentBottomOfImageMain() > bottomPercentInOption) {
                this._setPositionBottomOfImageMain(bottomPercentInOption, true);
            }
        },
        /**
         * Sets bottom position given in percentage (top css in fact).
         * @param {float}
         */
        _setPositionBottomOfImageMain: function (bottomValue, stop) {
            var topHeight = (this.options.maskOptions.top + this.options.maskOptions.height) * this.container.height / 100;
            this._setPositionTopOfImageMain(-((topHeight, this.imageMain.height - topHeight) / this.container.height * 100), stop);
        },
        /**
         * Sets left positon of image using given param.
         * @param {float} leftValue percentage value
         * @param {boolean} stop 
         */
        _setPositionLeftOfImageMain: function (leftValue, stop) {
            var rightPercentInOption = 0;
            this.imageMain.$element.css('left', leftValue + '%');

            if (stop === true) {
                return false;
            }

            if (this._getPercentLeftOfImageMain() > this.options.maskOptions.left) {
                this._setPositionLeftOfImageMain(this.options.maskOptions.left, true);
            }

            rightPercentInOption = (100 - (this.options.maskOptions.left + this.options.maskOptions.width));
            if (this._getPercentRightOfImageMain() > rightPercentInOption) {
                this._setPositionRightOfImageMain(rightPercentInOption, true);
            }
        },
        /**
         * Sets right position given in percentage (left css in fact).
         * @param {float}
         * @param {boolean} stop 
         */
        _setPositionRightOfImageMain: function (rightValue, stop) {
            var leftHeight = (this.options.maskOptions.left + this.options.maskOptions.width) * this.container.width / 100;
            this._setPositionLeftOfImageMain(-((leftHeight, this.imageMain.width - leftHeight) / this.container.width * 100), stop);
        },
        /**
         * Gets left percent of ImageMain.
         * @return {float} percent
         */
        _getPercentLeftOfImageMain: function () {
            return (this.imageMain.$element.position().left / this.container.width * 100);
        },
        /**
         * Gets right percent of ImageMain.
         * @return {float} percent
         */
        _getPercentRightOfImageMain: function () {
            return 100 - ((this.imageMain.$element.position().left + this.imageMain.width) / this.container.width * 100);
        },
        /**
         * Gets top percent of ImageMain.
         * @return {float} percent
         */
        _getPercentTopOfImageMain: function () {
            return (this.imageMain.$element.position().top / this.container.height * 100);
        },
        /**
         * Gets bottom percent of ImageMain.
         * @return {float} percent
         */
        _getPercentBottomOfImageMain: function () {
            return 100 - ((this.imageMain.$element.position().top + this.imageMain.height) / this.container.height * 100);
        },
        /**
         * Centers image vertically or horizontally based on its adjustedSize.
         * @return {void}
         */
        _centerImageMainInMask: function () {
            var self = this,
                options = this.options.maskOptions,
                opDimension = getOpositeDimension(this.imageMain.adjustedSize);

            var middle = (function () {
                var direction = (opDimension === 'width')  ? 'left' : 'top';
                return options[direction] + (options[opDimension] / 2);
            })();


            var diffPercentage = middle - (function halfSizeOfImage() {
                var halfSize = self.imageMain[opDimension] / 2,
                    percentOfMask = halfSize / self.imageMask[opDimension] * 100;

                return percentOfMask;
            })();

            if (this.imageMain.adjustedSize === 'width') {
                return this._setPositionTopOfImageMain(diffPercentage, true);
            }

            return this._setPositionLeftOfImageMain(diffPercentage, true);
        },
        /**
         * Adjust imageMain into Mask (which side should be movable).
         * @return {string} adjustedSize option of $imageMain
         */
        _adjustImageMainToMaskOrCropZone: function () {
            var self = this,
                options = this.options.maskOptions,
                imageRatio = this.imageMain.width / this.imageMain.height,
                maskOptionsRatio = (function getMaskOptionsRatio () {
                    if (self._isMaskDefinedInOptions() !== true) {
                        return options.width / options.height;
                    }
                    return (options.width * self.imageMask.width) / (options.height * self.imageMask.height);
                })();

            if (imageRatio - maskOptionsRatio > 0) {
                if (this._isMaskDefinedInOptions() === true) {
                    this.imageMain.$element.css({
                        'height': options.height + '%',
                        'top': options.top + '%',
                        'width': 'auto',
                        'left': 'auto'
                    });
                    this.imageMain.adjustedSize = 'height';
                }
                return this._updateCacheSizeOfImageMain();
            }
            if (this._isMaskDefinedInOptions() === true) {
                this.imageMain.$element.css({
                    'width': options.width + '%',
                    'left': options.left + '%',
                    'height': 'auto',
                    'top': 'auto'
                });
                this.imageMain.adjustedSize = 'width';
            }

            return this._updateCacheSizeOfImageMain();
        },
        /**
         * Removes inline styles from imageMain.
         * @return {void}
         */
        _removeImageMainStyles: function () {
            this.imageMain.$element.removeAttr('style');
        },
        /**
         * Removes imageMask from DOM.
         * @return {void}
         */
        _removeDomImageMask: function () {
            this.imageMask.$element.remove();
            this.imageMask = Object.create(elementObject);
        },
        /**
         * Removes cropZone from DOM.
         * @return {void}
         */
        _removeDomCropZone: function () {
            this.cropZone.$element.remove();
            this.cropZone = Object.create(elementObject);
        },
        /**
         * Appends imageMain into container.
         * @return {void}
         */
        _createDomImageMain: function () {
            var cClass = this.options.imageMainClass;
            this.imageMain.$element = $('<img/>').attr('src', this.options.imageMain || this.element.$element.attr('src')).addClass(cClass).appendTo(this.container.$element);
            this._updateCacheSizeOfImageMain();
        },
        /**
         * Appends cropZone into container.
         * @return {void}
         */
        _createDomCropZone: function () {
            var cClass = this.options.cropZoneClass;
            this.cropZone.$element = $('<div class="' + cClass +'"><div class="' + cClass + '-top' +'"></div><div class="' + cClass + '-right' +'"></div><div class="' + cClass + '-selection' +'"><span></span><span></span></div><div class="' + cClass + '-bottom' +'"></div><div class="' + cClass + '-left' +'"></div></div>').appendTo(this.container.$element);
        },
        /**
         * Appends imageMask into container.
         * @return {void}
         */
        _createDomImageMask: function () {
            var cClass = this.options.imageMaskClass;
            this.imageMask.$element = $('<img/>').attr('src', this.options.imageMask).addClass(cClass).appendTo(this.container.$element);
            this._updateCacheSizeOfImageMask();
        },
        /**
         * Sets CSS value of height.
         * @param  {string|number} value css height
         * @return {void}
         */
        _setContainerCssHeight: function (value) {
            this.container.$element.height(value);
            this._updateCacheSizeOfContainer();
        },
        /**
         * Update sizes of window
         * @return {void}
         */
        _updateCacheSizeOfWindow: function () {
            this.window.$element = $(window);
            this.window.width = this.window.$element.width();
            this.window.height = this.window.$element.height();
        },
        /**
         * Sets max-width of container based on window height.
         * @return {void}
         */
        _setMaxWidthOfContainer: function () {
            var imgRatio = 1,
                windowHeight = $(window).height(),
                diff = 0,
                maxWidth = 0;

            if ($.isNumeric(this.options.maxContainerMargin)) {
                if (this._isMaskDefinedInOptions()) {
                    imgRatio = this.imageMask.originalWidth / this.imageMask.originalHeight;
                }
                else {
                    imgRatio = this.imageMain.originalWidth / this.imageMain.originalHeight;
                }
                diff = windowHeight - this.container.$element.height();
                maxWidth = this.container.$element.width() + (diff * imgRatio);
                this.container.$element.css('max-width', Math.min(this.container.$element.parent().width(), maxWidth - this.options.maxContainerMargin) + 'px');
            }
        },
        /**
         * Gets size of given jquery object.
         * @param  {object} $element jquery object
         * @return {object} sizes of element: width and height
         */
        _getSizeOfGivenObject: function ($element) {
            return (function ($el) {
                return {
                    width: $el.width(),
                    height: $el.height()
                }
            })($element);
        },
        /**
         * Updates sizes of imageMain.
         * @return {void}
         */
        _updateCacheSizeOfImageMain: function () {
            var size = this._getSizeOfGivenObject(this.imageMain.$element);
            this.imageMain.width = size.width;
            this.imageMain.height = size.height;
        },
        /**
         * Updates sizes of container.
         * @return {void}
         */
        _updateCacheSizeOfContainer: function () {
            var size = this._getSizeOfGivenObject(this.container.$element);
            this.container.width = size.width;
            this.container.height = size.height;
        },
        /**
         * Updates sizes of imageMask.
         * @return {void}
         */
        _updateCacheSizeOfImageMask: function () {
            var size = this._getSizeOfGivenObject(this.imageMask.$element);
            this.imageMask.width = size.width;
            this.imageMask.height = size.height;  
        },
        /**
         * Updates sizes of imageMain.
         * @return {void}
         */
        _updateCacheSizeOfElement: function () {
            var size = this._getSizeOfGivenObject(this.element.$element);
            this.element.width = size.width;
            this.element.height = size.height;
        },
        /**
         * Checks is $element is visible
         * @return {boolean}
         */
        _isInitializingElementVisible: function () {
            return this.element.$element.is(':visible');
        },
        /**
         * Checks if cropZone is already in DOM.
         * @return {boolean}
         */
        _isCropZoneExistingInDom: function () {
            var cClass = this.options.cropZoneClass,
                isExisting = this.container.$element.find($('.' + cClass)).length > 0;

            if (isExisting && this.imageMain.$element === null) {
                this.cropZone.$element = this.container.$element.find($('.' + cClass));
            }

            return isExisting;
        },
        /**
         * Checks if imageMask is already in DOM.
         * @return {boolean}
         */
        _isImageMaskExistingInDom: function () {
            var cClass = this.options.imageMaskClass,
                isExisting = this.container.$element.find($('.' + cClass)).length > 0;

            if (isExisting && this.imageMask.$element === null) {
                this.imageMask.$element = this.container.$element.find($('.' + cClass));
                this._updateCacheSizeOfImageMask();
            }

            return isExisting;
        },
        /**
         * Checks if imageMain is already in DOM.
         * @return {boolean}
         */
        _isImageMainExistingInDom: function () {
            var cClass = this.options.imageMainClass,
                isExisting = this.container.$element.find($('.' + cClass)).length > 0;

            if (isExisting && this.imageMain.$element === null) {
                this.imageMain.$element = this.container.$element.find($('.' + cClass));
                this._updateCacheSizeOfImageMain();
            }

            return isExisting;
        },
        /**
         * Checks if container is already in DOM.
         * @return {boolean}
         */
        _isContainerExistingInDom: function () {
            var cClass = this.options.cropperContainerClass,
                isExisting = !this.options.checkContainerExistsInDom || this.element.$element.parent().find($('.' + cClass)).length > 0;

            if (isExisting && this.container.$element === null) {
                this.container.$element = this.element.$element.parents().not('picture').eq(0).find($('.' + cClass));
                this._updateCacheSizeOfContainer();
            }

            return isExisting;
        },
        /**
         * Checks if mask if defined in options.
         * @return {boolean}
         */
        _isMaskDefinedInOptions: function () {
            return typeof this.options.imageMask === 'string' && this.options.imageMask.length > 0;
        },
        /**
         * Setts Y pososition coords.
         * @param {float}
         * @return {void}
         */
        _setPositionCoordsY: function (yPosition) {
            var self = this;

            if ($.isNumeric(yPosition)) {
                if (this._isMaskDefinedInOptions()) {
                    return;
                }
                this.options.positionCoords.y = Math.max(Math.min(yPosition, 100 - self.options.maskOptions.height), 0.0);
            }
        },
        /**
         * Setts X pososition coords.
         * @param {float}
         * @return {void}
         */
        _setPositionCoordsX: function (xPosition) {
            var self = this;

            if ($.isNumeric(xPosition)) {
                if (this._isMaskDefinedInOptions()) {
                    return;
                }
                this.options.positionCoords.x = Math.max(Math.min(xPosition, 100 - self.options.maskOptions.width), 0.0);
            }
        },
        _setMaskOptionsHeight: function (height) {
            return this.options.maskOptions.height = Math.min(Math.max(height, 0.0), 100);
        },
        _setMaskOptionsWidth: function (width) {
            return this.options.maskOptions.width = Math.min(Math.max(width, 0.0), 100);
        }
    };

    /**
     * Plugin constructor.
     * 
     * @param  {object} options plugin options
     * @return {object}
     */
    $.fn.imageMaskCropper = function (options) {
        this.maskCropperInstance = null;

        var self = this,
            settings = $.extend(true, $.fn.imageMaskCropper.defaults, (function() {
                try {
                    var imageMask = self.attr('data-mask-image'),
                        imageMain = self[0].currentSrc || self.attr('src'),
                        result = {
                            imageMask: imageMask,
                            imageMain: imageMain
                        };

                    if (self.attr('data-mask-options')) {
                        result.maskOptions = $.parseJSON(self.attr('data-mask-options'));
                    }
                    if (self.attr('data-position-coords')) {
                        result.positionCoords = $.parseJSON(self.attr('data-position-coords'));
                    }
                } catch (e) {
                    return {};
                }
                return result;
            })(), options);

        self.maskCropperInstance = new imageMaskCropper(settings, self);
        return this.maskCropperInstance;
    };
    /**
     * Default options of plugin.
     * 
     * @type {object}
     */
    $.fn.imageMaskCropper.defaults = {
        imageMask: '',
        imageMain: '',
        /**
         * PercentageValues.
         * (left, top) properties available only for ImageMask.
         * @type {object}
         */
        maskOptions: {
            left: 0,
            top: 0, 
            width: 100,
            height: 100
        },
        /**
         * Initial coords.
         * @type {boolean|object}
         */
        positionCoords: {
            x: 0,
            y: 0
        },
        enableCrop: true,
        checkContainerExistsInDom: true,
        maxContainerMargin: false,
        cropperContainerClass: 'image-mask-cropper',
        cropperContainerClassHasMask: 'has-mask',
        cropperContainerClassLoading: 'images-loading',
        cropperContainerClassEnabledCrop: 'enabled-crop',
        cropZoneClass: 'cropper-crop-zone',
        imageMainClass: 'cropper-image-main',
        imageMaskClass: 'cropper-image-mask',
        isMobile: false,
        onInitialized: function() {} || false,
        onMoveCrop: function() {} || false
    };
    /* <--END: plugin core-->  */
}));
